// module hu_rs232.cpp
// Simple serial (RS232) communication
//
// Simple protocol used:
//   first package: file name 
//   n content packages 
//   last package: length 0 
//
//   each package is preceeded by a length (6 ASCII numeric bytes) 
//   each package (the length also) has a checksum as the final byte
//   each package (the length also) must be confirmed by a single ack byte
//
// Dr. E.Huckert, Goldbach(Germany) 2002-04-20
// Copyright Dr.E.Huckert 04-2002

// 2017-01-05 added method getErrorString()
//            method read_line() redesigned
//            important change for COM lines >= COM10

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <io.h>
#include "hu_rs232.h"

#define CREAT_MASK   0xff       // nur fuer MSDOS

// ------------------------------------------------------------
// Constructor
HU_RS232::HU_RS232(unsigned cno,
                   unsigned long baud)
{
  this->writeflag   = 0;
  this->baudrate    = baud;
  this->comno       = cno;
  this->comm_handle = INVALID_HANDLE_VALUE;
  this->logging     = 0;
  this->sum_retrans = 0;
  this->loc_retrans = 0;
  this->maxtimeout  = 100;
}

// ---------------------------------------------------------
// destructor
HU_RS232::~HU_RS232()
{
}   

// -------------------------------------------------------------------
// Nur fuer Windows Fehler
const char * HU_RS232::getErrorString()
{
  DWORD dwError      = ::GetLastError();
  static LPVOID pTextBuffer = NULL;
  static char noErrorMsg[1] = 0;
  if (NOERROR == dwError)
  {
    noErrorMsg[0] = 0;
    return noErrorMsg;
  }
  const DWORD dwFormatControl = FORMAT_MESSAGE_ALLOCATE_BUFFER |
                                  FORMAT_MESSAGE_IGNORE_INSERTS |
                                  FORMAT_MESSAGE_FROM_SYSTEM;
  DWORD dwCount = ::FormatMessage(dwFormatControl, NULL, 
                                   dwError, 0, 
                                  (LPTSTR) &pTextBuffer, 
                                  0, NULL);
  return (const char *)pTextBuffer;
}   // end HU_RS232::getErrorString()

// ---------------------------------------------------------
// Returns a simple checksum
unsigned char HU_RS232::checksum(char *buf,
                                 int lbuf)
{
int n;
unsigned char cs;
  cs = 0;
  for (n=0; n < lbuf; n++)
    cs = cs + (unsigned char)(buf[n]);
  return cs;
}   // end HU_RS232::checksum()

// ---------------------------------------------------------
// Returns: >= 0 if ok
//           < 0 if error
// side effect: comm_handle set 
int HU_RS232::open_line(int  writeflag)
{
int    ret = 0;
char   hbuf[256];
char   comm_device[64];
HANDLE handle = 0;
DWORD  mask;
DCB    dcb;
  if (this->comno < 0)
  {
    //::sprintf(hbuf, "Fehler open_line(): comno < 0!\n");
    ret = -1;
    goto zurueck;
  }
  if (this->comno < 10)
    ::sprintf(comm_device, "COM%u", this->comno);
  else
  {
    // s.  Link: https://support.microsoft.com/en-us/kb/115831
    // COM ports >=10 must be given as (sample): "\\.\COM10"  ((double \ needed for C/C++)
    /*
            "The format for the COM port specifier for CreateFile() for COM ports greater than
            COM9 appear to need the \\.\ as a prefix. See HOWTO: Specify Serial Ports Larger than 
               COM9 from Microsoft Support."
         */
    char aux[32];
    ::sprintf(aux, "COM%u", this->comno);
    ::strcpy(comm_device, "\\\\.\\");
    ::strcat(comm_device, aux);
  }
  if (writeflag)
  {
    handle = ::CreateFile(TEXT(comm_device), 
                          GENERIC_READ + GENERIC_WRITE,
                          0, 
                          NULL,
                          OPEN_EXISTING, 
                          0, // FILE_ATTRIBUTE_NORMAL, 
                          NULL);
  }
  else
  {
    handle = ::CreateFile(comm_device, 
                          GENERIC_READ + GENERIC_WRITE,
                          0, 
                          NULL,
                          OPEN_EXISTING, 
                          0,    // FILE_ATTRIBUTE_NORMAL, 
                          NULL);
  }
  if (handle == INVALID_HANDLE_VALUE)
  {
    ret = -2;
    goto zurueck;
  }
  // 
  // the line has been successfully opened
  this->comm_handle = handle;
  //
  memset((void *)&dcb, 0, sizeof(dcb));
  dcb.BaudRate      = (DWORD)baudrate;
  dcb.fBinary       = 1;
  dcb.fDtrControl   = DTR_CONTROL_DISABLE;
  dcb.fRtsControl   = RTS_CONTROL_DISABLE;
  dcb.fInX          = FALSE;
  dcb.fOutX         = FALSE;
  dcb.Parity        = NOPARITY;
  dcb.StopBits      = ONESTOPBIT;
  dcb.ByteSize      = 8;
  SetCommState(comm_handle, &dcb);
  //
  mask = EV_RXCHAR;
  //SetCommMask(comm_handle, mask);
  /* from MS docu:
        A value of MAXDWORD, combined with zero values for both the ReadTotalTimeoutConstant 
         and ReadTotalTimeoutMultiplier members, specifies that the read operation is to return 
         immediately with the bytes that have already been received, even if no bytes have been received.
      */
  COMMTIMEOUTS  CommTimeOuts ;
  CommTimeOuts.ReadIntervalTimeout         = MAXDWORD;
  CommTimeOuts.ReadTotalTimeoutMultiplier  = 0;
  CommTimeOuts.ReadTotalTimeoutConstant    = 0;
  CommTimeOuts.WriteTotalTimeoutMultiplier = 1;
  CommTimeOuts.WriteTotalTimeoutConstant   = 2;
  ret = 1;
  if (! SetCommTimeouts(comm_handle, &CommTimeOuts))
  {
    ret = -3;
    goto zurueck;
  }
  //
  zurueck:  
  if (this->logging)
    printf("HU_RS232::open_line() ret=%d\n", ret);
  return ret;
}   // end HU_RS232::open_line()

// ---------------------------------------------------------
// Read a maximum of lbuf bytes from the line
// Input lines should be terminated by LF
// Returns: the number of bytes actually read
// Not for binary data!!!!! Reacts on LF!
// Returns < 0 if error or timeout
int HU_RS232::read_line(char *buf, int lbuf)
{
DWORD    nb;
BOOL     ret;
int      funcRet = 0;
int      ll, pos, anz;
char     ch;
unsigned timeout;
DWORD    mask;
  //
  if (logging >1)
    printf("read_line(): %d bytes\n", lbuf);
  if (comm_handle == INVALID_HANDLE_VALUE) 
  {
    funcRet = -1;
    goto zurueck;
  }
  if (logging > 1)
  {
    ::GetCommMask(comm_handle, &mask);
    printf("GetCommMask():%08lx\n", mask);
    ::GetCommModemStatus(comm_handle, &mask);
    printf("GetCommModemStatus():%08lx\n", mask);
  }
  ll      = lbuf;
  pos     = 0;
  timeout = 0;
  //
  while (ll > 0)
  { 
    anz = 1;  // received only one byte
    if (anz > ll) 
      anz = ll;
    ret = ::ReadFile(comm_handle, 
                     (LPSTR)&buf[pos], 
                     (DWORD)anz, 
                     &nb, 
                     NULL);  
    if (ret == FALSE)
    {
      printf("Read ERROR on the serial line\n");
      funcRet = -2;
      goto zurueck;
    }
    if (logging > 1)
      printf("\t received: %ld bytes\n", nb);
    if (nb == 0L)
    {
      ::Sleep(1L);
      timeout++;
      if (timeout > this->maxtimeout)
      {
        if (pos > 0)
          // already received something
         break;
        // nothing received here
        if (logging)
          printf("ERROR: receive timeout\n");
        funcRet = 0;
        goto zurueck;
      }
      continue;
    }  // end if (nb == 0L)
    else
      timeout = 0;
    ch = buf[pos];
    pos += nb;
    ll  -= nb;
    if (ch == '\n')
      // end-of-line received
      break;
  }   // end while ...
  //
  funcRet = pos;
  //
  zurueck:
  if (logging > 1)
    printf(" HU_RS232::read_line() funcRet=%d\n", funcRet);
  return funcRet;
}   // end HU_RS232::read_line()

// ---------------------------------------------------------
// wait for a single character as the acknowledgement
// Returns: the ack character
//          < 0 if error
int HU_RS232::wait_ack()
{
char buf[2];
int ret;
  if (logging > 1)
    printf("wait_ack()\n");
  ret = read_line(buf, 1);
  if (ret < 0) 
    return ret;
  else
    return (int)(buf[0]);
}   // end HU_RS232::wait_ack()

// ---------------------------------------------------------
// write a buffer on the serial line (must be open)
// Returns: <0  if error
//        : >=0 if ok
int HU_RS232::put_line(char *buf, 
                       int  lbuf)
{
DWORD    nb;
BOOL     ret;
int      ll, pos, anz;
  if (logging)
    printf("put_line(): %d bytes\n", lbuf);
  if (comm_handle == INVALID_HANDLE_VALUE)
    return -1;
  ll  = lbuf;
  pos = 0;
  // angepasst fuer Versuche mit Rexroth LKs!!!!
  while (ll > 0)
  {  
    ::Sleep(1);  // fuer Rexroth Lesekoepfe - sonst nicht noetig!
    anz = 1;
    if (anz > ll) anz = ll;
    ret = WriteFile(comm_handle, (LPSTR)&buf[pos], (DWORD)anz, &nb, NULL);
    if (ret == FALSE)
    {
      printf("Write ERROR on the serial line\n");
      return -2;
    }
    ret = ::FlushFileBuffers(comm_handle);
    if (logging > 1)
    {
      printf("\tsent %ld bytes\n", nb);
    }
    pos += nb;
    ll  -= nb;
  }   // end while (1)
  //
  if (logging)
    printf("result put_line(): %d\n", lbuf);
  return lbuf;
}   // end HU_RS232::put_line()

// ---------------------------------------------------------
// Returns: < 0 if error
int HU_RS232::close_line()
{
  if (this->comm_handle == INVALID_HANDLE_VALUE)
    return -1;
  CloseHandle(this->comm_handle);
  this->comm_handle = INVALID_HANDLE_VALUE;
  return 0;
}   // end HU_RS232::close_line()

// ---------------------------------------------------------
int HU_RS232::send_ack(char ack)
{
int ret;
char buf[2];
  if (logging)
    printf("send_ack(): %c\n", ack);
  buf[0] = ack;    // '1' means OK
  ret = this->put_line(buf, 1);
  return ret;
}   // send send_ack()

// ---------------------------------------------------------
// Receive a file over the serial line
// returns: 0  if ok
//        : <0 if error
int HU_RS232::receive_file(char *fn)
{
char *buf;
char xbuf[7];
unsigned char cs;
int  ofd, ret, anz, nopack;
  ofd = creat(fn, CREAT_MASK);
  if (ofd < 0)
    return -1;
  setmode(ofd, O_BINARY);
  //
  // receive the first package: the file name
  // we don't use this file name at the moment
  this->loc_retrans = 0;
again1:
  this->loc_retrans++;
  ret = this->read_line(xbuf, 7);
  if (ret < 0)
    return -2;
  cs = this->checksum(xbuf, 6);
  if (cs != (unsigned char)(xbuf[6]))
  {
    if (this->loc_retrans > 10)
    {
      ret = send_ack(0xff);
      return -22;
    }
    ret = send_ack('0');
    goto again1;
  }
  ret = send_ack('1');
  xbuf[6] = 0;
  anz = atoi(xbuf);
  buf = new char[anz + 1];
  this->loc_retrans = 0;
again2:
  this->loc_retrans++;
  ret = this->read_line(buf, anz + 1);
  if (ret < 0)
    return -3;
  cs = this->checksum(buf, anz);
  if (cs != (unsigned char)(buf[anz]))
  {
    if (this->loc_retrans > 10)
    {
      ret = send_ack(0xff);
      return -33;
    }
    ret = send_ack('0');
    goto again2;
  }
  buf[anz] = 0;
  delete buf;
  ret = send_ack('1');
  if (ret < 0)
    return -3;
  if (logging)
    printf("filename %s received\n", buf);
  //
  // receive the content packages
  //
  nopack = 0;
  while (1)
  {
    this->loc_retrans = 0;
again3:
    this->loc_retrans++;
    // read the length of the next package
    ret = this->read_line(xbuf, 7);
    if (ret < 0)
      return -4;
    cs = checksum(xbuf, 6);
    if (cs == (unsigned char)(xbuf[6]))
      ret = send_ack('1');
    else
    {
      // wrong checksum: request retransmission
      if (this->loc_retrans > 10)
      {
        ret = send_ack(0xff);
        return -44;
      }
      ret = send_ack('0');
      goto again3;
    }
    xbuf[6] = 0;
    anz = atoi(xbuf);
    if (anz == 0)
      // this was the final package
      break;
    buf = new char[anz + 1];
    this->loc_retrans = 0;
again4:
    this->loc_retrans++;
    //
    // read the content of the next package
    ret = this->read_line(buf, anz + 1);
    if (ret < 0)
      return -5;
    //
    // acknowledge the last package
    cs = checksum(buf, anz);
    if (cs != (unsigned char)(buf[anz]))
    {
      // wrong checksum: retransmit
      if (this->loc_retrans > 10)
      {
        ret = send_ack(0xff);
        return -55;
      }
      ret = send_ack('0');
      goto again4;
    }
    //
    // the checksum was good
    ret = write(ofd, buf, anz);
    if (ret < 0)
      return -7;
    ret = send_ack('1');   // positive ack
    delete buf;
    nopack++;
    if (logging)
      printf("package %d received\n", nopack);
  }   // end HU_RS232::while 1
  //
  close(ofd); 
  return 0;
}   // end HU_RS232::receive_file()

// ---------------------------------------------------------
// returns: 0  if ok
//        : <0 if error
int HU_RS232::send_file(char *fn)
{
int  anz, ifd, ll, nb, ret, nopack;
char xbuf[16];
char buf[513];
unsigned char cs;
  ifd = open(fn, 0);
  if (ifd < 0)
    return -1;
  setmode(ifd, O_BINARY);
  //
  // send the filename as the first package 
again1:
  sprintf(buf, "%06u", strlen(fn));
  cs = this->checksum(buf, 6);
  buf[6] = cs;
  ret = this->put_line(buf, 7);   // length of the filename
  if (ret < 0) 
    return -2;
  ret = this->wait_ack();
  if (ret != '1') 
  {
    if ((ret & 0xff) == 0xff)
      return -22;
    // retransmission necessary
    goto again1;
  }
  //
again2:
  ll = strlen(fn);
  strcpy(buf, fn);
  cs = this->checksum(buf, ll);
  buf[ll] = cs;
  ret = this->put_line(buf, ll + 1);   // the filename
  if (ret < 0) 
    return -3;
  ret = this->wait_ack();
  if (ret != '1') 
  {
    if ((ret & 0xff) == 0xff)
      return -33;
    // retransmission necessary
    goto again2;
  }
  if (logging)
    printf("filename %s sent\n", fn);
  //
  nopack = 0;
  while (1)
  {
    // send the length of the package
    anz = sizeof(buf) - 1;
    nb = read(ifd, buf, anz);
    if (nb <= 0)
      break;
again3:
    //
    // send the length of the package
    sprintf(xbuf, "%06u", nb);
    cs = this->checksum(xbuf, 6);
    xbuf[6] = cs;
    ret = this->put_line(xbuf, 7); 
    if (ret < 0) 
      return -4;
    ret = this->wait_ack();
    if (ret != '1') 
    {
      if ((ret & 0xff) == 0xff)
        return -44;
      // retransmission necessary
      goto again3;
    }
    //
    // send the content
again4:
    cs = this->checksum(buf, nb);
    buf[nb] = cs;
    ret = this->put_line(buf, nb + 1); 
    if (ret < 0) 
      return -5;
    //
    // wait for the acknowledgement
    ret = this->wait_ack();
    if (ret != '1') 
    {
      if ((ret & 0xff) == 0xff)
        return -44;
      // retransmission necessary
      goto again4;
    }
    //
    nopack++;
    if (logging)
      printf("package %d sent\n", nopack);
  }
  //
  // send the final (last) package
again5:
  strcpy(xbuf, "000000");
  cs = this->checksum(xbuf, 6);
  xbuf[6] = cs;
  ret = this->put_line(xbuf, 7);
  ret = this->wait_ack();
  if (ret != '1') 
  {
    if ((ret & 0xff) == 0xff)
      return -55;
    // retransmission necessary
    goto again5;
  }
  //
  return 0;    // OK
}   // end HU_RS232::send_file()
